#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import uuid
import json
import time

from paho.mqtt.client import Client

import quasimodo.base


class QueueWorkerSkeletonTT(quasimodo.base.Q):
    DEFAULT_HOST = "localhost"
    DEFAULT_PORT = 15675
    DEFAULT_USERNAME = None
    DEFAULT_PASSWORD = None
    DEFAULT_HEARTBEAT_INTERVAL = 60
    DEFAULT_ENDPOINT = "/ws"
    DEFAULT_BINDING_KEYS = ["*", "*.*"]

    def __init__(self, *args, **kwargs):
        quasimodo.base.Q.__init__(self, *args, **kwargs)
        self.log = logging.getLogger(__name__)

        self.endpoint = kwargs.get(
            "endpoint", getattr(self, "DEFAULT_ENDPOINT")
        )
        self.heartbeat_interval = kwargs.get("heartbeat_interval", 60)
        client_id = kwargs.get("client_id")

        if client_id is None:
            client_id = str(uuid.uuid4())

        self.exchange_binding_keys = kwargs.get("binding_keys")

        self.connected = False
        if self.transport is None:
            self.transport = "websockets"
        self.client = Client(client_id=client_id, transport=self.transport)
        self.client.ws_set_options(self.endpoint)
        self.client.on_connect = self.__on_connect
        self.client.on_message = self.callback
        self.client.on_subscribe = self.__on_subscribe

        self.client.username_pw_set(self.credentials[0], self.credentials[1])

        self.client.enable_logger()

        if self.tls_context:
            self.client.tls_set_context(self.tls_context)

        if self.autorun:
            self.run()

    def __on_connect(self, client, userdata, flags, rc):
        self.log.debug("on connect ... (rc={!r})".format(rc))
        self.connected = rc == 0

        if self.connected:
            if not self.exchange_binding_keys:
                self.exchange_binding_keys = self.DEFAULT_BINDING_KEYS
                self.log.warning(
                    "Using default binding keys {!r}".format(
                        self.exchange_binding_keys
                    )
                )

            for binding_key in self.exchange_binding_keys:
                client.subscribe(binding_key)

    def callback(self, client, userdata, message):
        try:
            payload = json.loads(message.payload)
        except Exception as exc:
            print(repr(message.payload))
            return

        self.handle_request(
            payload, client=client, userdata=userdata, message=message
        )

    def __on_subscribe(self, client, userdata, mid, granted_qos):
        self.log.debug("on subscribe ...")
        # print(client)
        # print(userdata)
        # print(mid)
        # print(granted_qos)

    def run(self):
        self.client.connect(
            self.host, self.port, keepalive=self.heartbeat_interval
        )
        self.client.loop_start()

        self._start_consuming()

    def _start_consuming(self):
        net_loc = "{username!s}:{password!s}@{host}:{port}".format(
            username=self.client._username,
            password=self.client._password,
            host=self.host,
            port=self.port,
        )

        exchange_binding_keys = self.exchange_binding_keys
        if exchange_binding_keys is None:
            exchange_binding_keys = self.DEFAULT_BINDING_KEYS

        path_indicator = ""
        if self.transport == "websockets":
            path_indicator = f"{self.client._websocket_path} "
        listening_to = "{path_indicator}({binding_keys})".format(
            path_indicator=path_indicator,
            binding_keys="; ".join(sorted(exchange_binding_keys)),
        )

        self.log.debug(
            "The monkeys are listening to {:s} {:s}".format(
                net_loc, listening_to
            )
        )

        while True:
            time.sleep(0.5)

    def simple_publish(self, payload, routing_key="", **kwargs):
        sp_client_id = "spc-" + str(uuid.uuid4())
        sp_client = Client(client_id=sp_client_id, transport=self.transport)
        sp_client.ws_set_options(self.endpoint)

        if self.tls_context:
            sp_client.tls_set_context(self.tls_context)
        sp_client.username_pw_set(self.credentials[0], self.credentials[1])
        sp_client.connect(
            self.host, self.port, keepalive=self.heartbeat_interval
        )

        sp_client.loop_start()

        res = sp_client.publish(routing_key, payload=json.dumps(payload))
        res.wait_for_publish()

        sp_client.loop_stop()
        self.log.debug("rc={!r}".format(res.rc))
        return res.rc == 0

    def _handle_request(self, payload, **kwargs):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import uuid
import json
import pprint
import copy

import pika

import quasimodo
import quasimodo.base

MIMETYPE_JSON = "application/json"

MIMETYPE_BINARY = "application/octet-stream"

#: fallback queue
DEFAULT_QUEUE = "quasimodo"


class Quasimodo(quasimodo.base.Q):
    #: default rabbitMQ parameters
    DEFAULT_HOST = "localhost"
    DEFAULT_PORT = 5671
    DEFAULT_USERNAME = "guest"
    DEFAULT_PASSWORD = "guest"
    DEFAULT_HEARTBEAT_INTERVAL = 1800
    DEFAULT_BINDING_KEYS = ["*"]

    def __init__(self, *args, **kwargs):
        quasimodo.base.Q.__init__(self, *args, **kwargs)
        self.log = logging.getLogger(__name__)
        self.ssl_options = None
        if self.tls_context:
            self.ssl_options = pika.SSLOptions(self.tls_context, self.host)

    def set_credentials(self, username, password):
        self._credentials = pika.PlainCredentials(
            username=username, password=password
        )

    def add_to_queue(
        self,
        payload,
        queue=DEFAULT_QUEUE,
        content_type=MIMETYPE_JSON,
        arguments=None,
        correlation_id=None,
        reply_to=None,
        durable=False,
    ):
        """
        Request adding of *payload* to the queue named *queue*.
        """
        transaction_id = uuid.uuid4().hex

        payload["transaction_id"] = transaction_id
        payload["queue"] = queue

        if content_type == MIMETYPE_JSON:
            payload = json.dumps(payload)

        self.log.debug("Payload (q={!r}):".format(queue))
        self.log.debug(pprint.pformat(payload))

        connection_parameters = pika.ConnectionParameters(
            host=self.host,
            port=self.port,
            credentials=self.credentials,
            ssl_options=self.ssl_options,
        )
        connection = pika.BlockingConnection(connection_parameters)
        channel = connection.channel()

        properties = pika.BasicProperties(
            # make message persistent
            delivery_mode=2,
            content_type=content_type,
            correlation_id=correlation_id,
            reply_to=reply_to,
        )

        if arguments:
            channel.queue_declare(
                queue=queue, durable=True, arguments=arguments
            )
        elif durable:
            channel.queue_declare(queue=queue, durable=True)
        else:
            channel.queue_declare(queue=queue, passive=True)

        channel.basic_publish(
            exchange="", routing_key=queue, body=payload, properties=properties
        )

        connection.close()

        return transaction_id

    def _add_to_exchange(
        self,
        payload,
        routing_key="",
        exchange="amq.topic",
        content_type=MIMETYPE_BINARY,
        **kwargs,
    ):
        """
        Request adding of *payload* to the exchange named *exchange*.
        """
        if not kwargs.get("properties"):
            properties = pika.BasicProperties(
                # make message persistent
                delivery_mode=2,
                content_type=content_type,
            )
        else:
            properties = kwargs.get("properties")

        self.channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=payload,
            properties=properties,
        )

    def add_to_exchange(
        self,
        payload,
        routing_key="",
        exchange="amq.topic",
        content_type=MIMETYPE_JSON,
        exchange_type="direct",
        arguments=None,
    ):
        """
        Request adding of *payload* to the exchange named *exchange*.
        """
        transaction_id = uuid.uuid4().hex

        payload["transaction_id"] = transaction_id
        payload["exchange"] = exchange
        if content_type == MIMETYPE_JSON:
            payload = json.dumps(payload)

        self.log.debug("Payload (x={!r}):".format(exchange))
        self.log.debug(pprint.pformat(payload))

        connection_parameters = pika.ConnectionParameters(
            host=self.host,
            port=self.port,
            credentials=self.credentials,
            ssl_options=self.ssl_options,
        )
        connection = pika.BlockingConnection(connection_parameters)
        channel = connection.channel()

        properties = pika.BasicProperties(
            # make message persistent
            delivery_mode=2,
            content_type=content_type,
        )

        if arguments:
            channel.exchange_declare(
                exchange=exchange,
                durable=True,
                arguments=arguments,
                exchange_type=exchange_type,
            )
        else:
            channel.exchange_declare(
                exchange=exchange, passive=True, exchange_type=exchange_type
            )

        channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=payload,
            properties=properties,
        )

        connection.close()

        return transaction_id

    @property
    def qcc(self):
        return self

    def simple_publish(self, payload, routing_key="", **kwargs):
        return self.add_to_exchange(payload, routing_key=routing_key, **kwargs)


class QueueWorkerSkeleton(Quasimodo):
    """
    Skeleton implementation of a queue worker application.
    """

    def __init__(self, *args, **kwargs):
        Quasimodo.__init__(self, *args, **kwargs)
        self.log = logging.getLogger(kwargs.get("log_name", __name__))
        self.queue_name = kwargs.get("queue", False)
        self.exchange_name = kwargs.get("exchange", False)
        self.exchange_type = kwargs.get("exchange_type", "topic")
        self.exchange_binding_keys = kwargs.get("binding_keys")
        self.requeue_default = kwargs.get("requeue_default", False)
        self.max_consumed_messages = kwargs.get("max_consumed_messages", -1)
        self.deadletter_support = kwargs.get("deadletter_support", False)
        self.queue_declare_arguments = kwargs.get("queue_declare_arguments")
        self.exchange_queue_declare_arguments = kwargs.get(
            "exchange_queue_declare_arguments"
        )
        self.consumed_count = 0
        self.channel = None
        self.connection = None

        if self.autorun:
            self.run()

    def run(self):
        """
        Set up exchange/queue (blocking) and start consuming.
        """
        connection_parameters = pika.ConnectionParameters(
            host=self.host,
            port=self.port,
            credentials=self.credentials,
            heartbeat=self.heartbeat_interval,
            connection_attempts=10,
            ssl_options=self.ssl_options,
        )
        self.connection = pika.BlockingConnection(connection_parameters)
        self.channel = self.connection.channel()

        if self.exchange_name is not False:
            if not self.exchange_binding_keys:
                self.exchange_binding_keys = self.DEFAULT_BINDING_KEYS
                self.log.warning(
                    f"Using default binding keys {self.exchange_binding_keys!r}"
                )

            self.channel.exchange_declare(
                exchange=self.exchange_name,
                exchange_type=self.exchange_type,
                arguments=self.queue_declare_arguments,
                durable=True,
            )

            result = self.channel.queue_declare(
                "",
                exclusive=True,
                arguments=self.exchange_queue_declare_arguments,
            )
            self.queue_name = result.method.queue

            for binding_key in self.exchange_binding_keys:
                self.channel.queue_bind(
                    exchange=self.exchange_name,
                    queue=self.queue_name,
                    routing_key=binding_key,
                )
        else:
            if self.deadletter_support:
                self._setup_deadletter_exchange()
            else:
                self.channel.queue_declare(
                    queue=self.queue_name,
                    durable=True,
                    arguments=self.queue_declare_arguments,
                )

        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(
            on_message_callback=self.callback, queue=self.queue_name
        )

        self._start_consuming()

    def _setup_deadletter_exchange(self):
        """
        Add dead letter exchange handling if enabled.
        """
        if not self.deadletter_support:
            self.log.debug("No dead letter support")
            return

        exchange_name = "{:s}-dlx".format(self.queue_name)
        deadletter_name = "{:s}-deadletter".format(self.queue_name)

        self.log.info("Setting up exchange {:s}".format(exchange_name))
        self.channel.exchange_declare(
            exchange=exchange_name, exchange_type="direct", durable=True
        )

        self.log.debug(
            "Setting up deadletter queue {:s}".format(deadletter_name)
        )
        self.channel.queue_declare(queue=deadletter_name, durable=True)

        arguments = {
            # 'x-message-ttl': 1000,
            "x-dead-letter-exchange": exchange_name,
            # if not specified, queue's routing-key is used
            "x-dead-letter-routing-key": self.queue_name,
        }
        self.channel.queue_declare(
            queue=self.queue_name, durable=True, arguments=arguments
        )

    def _start_consuming(self):
        """
        Start consuming function.
        May be implemented by deriving classes.
        """
        net_loc = "{username!s}:{password!s}@{host}:{port}".format(
            username=self._credentials.username,
            password=self._credentials.password,
            host=self.host,
            port=self.port,
        )
        listening_to = "{queue}".format(queue=self.queue_name)
        if self.exchange_name is not False:
            listening_to += " ({:s})".format(
                "; ".join(sorted(self.exchange_binding_keys))
            )
        self.log.debug(
            "The monkeys are listening to {:s} {:s}".format(
                net_loc, listening_to
            )
        )

        self.channel.start_consuming()

    def callback(self, channel, method, properties, body):
        """
        Message queue callback.
        """
        may_run = True
        success = None
        requeue = self.requeue_default
        self.consumed_count += 1
        response_payload = dict()
        payload = dict()
        handle_request_result = dict()

        if properties.content_type == MIMETYPE_JSON:
            body = json.loads(body)

        if properties.content_type is None:
            try:
                body = json.loads(body)
                properties.content_type = MIMETYPE_JSON
            except Exception:
                pass

        if self.verbose:
            self.log.info("Received    S{:s}".format(pprint.pformat(body)))
            self.log.info(
                " properties {:s}".format(pprint.pformat(properties))
            )
            self.log.info(" channel    {:s}".format(pprint.pformat(channel)))
            self.log.info(" method     {:s}".format(pprint.pformat(method)))

        if not may_run:
            self.log.warning("Denied by may_run value")
            requeue = False
        else:
            payload = copy.copy(body)
            try:
                del payload["transaction_id"]
            except Exception:
                pass

            try:
                self.log.debug("now handling request")
                handle_request_result = self.handle_request(
                    payload=payload,
                    channel=channel,
                    method=method,
                    properties=properties,
                )
            except Exception as exception:
                self.log_traceback(
                    "execution (QueueWorkerSkeleton.callback)", exception
                )

        try:
            success = handle_request_result.get("success")
            response_payload = handle_request_result
        except AttributeError:
            success = handle_request_result
            response_payload["success"] = handle_request_result
        except NameError as nexc:
            self.log_traceback("UH-OH! execution (QueueWorkerSkeleton)", nexc)
            success = False
            requeue = False

        response_payload["__request_payload"] = payload

        self._confirm_callback(channel, method, properties, response_payload)

        if success:
            self.log.debug("Success.")
            channel.basic_ack(delivery_tag=method.delivery_tag)
        else:
            self.log.warning(
                "{!s}: Failure, Abort. requeue={!r}".format(
                    self.identifier, requeue
                )
            )
            self.log.info("payload:")
            self.log.info("-" * 40)
            self.log_data_dump(body)
            self.log.info("~" * 40)

            if not self.deadletter_support:
                channel.basic_nack(
                    delivery_tag=method.delivery_tag, requeue=requeue
                )
            else:
                delivery_tag = method.delivery_tag
                requeue = False
                self.log.info(
                    "basic_reject: delivery_tag={delivery_tag} requeue={requeue}".format(
                        delivery_tag=delivery_tag, requeue=requeue
                    )
                )
                channel.basic_reject(
                    delivery_tag=delivery_tag, requeue=requeue
                )

        if self.max_consumed_messages != -1:
            if self.consumed_count >= self.max_consumed_messages:
                try:
                    current_identifier = " ({:s})".format(self.identifier)
                except Exception:
                    current_identifier = ""
                self.log.info("Stop consuming{:s}".format(current_identifier))
                channel.stop_consuming()

    def _confirm_callback(self, channel, method, properties, payload):
        """
        Callback for emitting confirmations.
        """
        self.log.debug("CONFIRM CALLBACK:")
        try:
            if not properties.reply_to:
                self.log.debug("reply_to is None")
                return None

            mfg = "[{reply_to}] corr'n_id={correlation_id}: {payload}".format(
                reply_to=properties.reply_to,
                correlation_id=properties.correlation_id,
                payload=pprint.pformat(payload),
            )
            self.log.info(mfg)

            pika_properties = pika.BasicProperties(
                correlation_id=properties.correlation_id,
                content_type=MIMETYPE_JSON,
            )

            try:
                payload_json = json.dumps(payload)
            except TypeError as texception:
                payload_dummy = {
                    "_exception": True,
                    "_payload": pprint.pformat(payload),
                }
                self.log_traceback("JSON conversion failed", texception)
                payload_json = json.dumps(payload_dummy)

            channel.basic_publish(
                exchange="",
                routing_key=properties.reply_to,
                properties=pika_properties,
                body=payload_json,
            )
            return True
        except Exception as exception:
            self.log_traceback("confirm callback", exception)

        self.log.warn("FAIL!")

        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import argparse
import logging
import json
import socket

from quasimodo.stubs import Ape, Monkey

OPT_NO_VALUE = "<use implementation's default>"


class DumpMixin(object):
    def _handle_request(self, payload, **kwargs):
        success = True

        self.log_data_dump(payload)

        return {"success": success}


class ApeDumper(Ape, DumpMixin):
    pass


class MonkeyDumper(Monkey, DumpMixin):
    pass


def dump_hunchback_parameters(kwargs, connection_option_keys, use_log=None):
    if use_log is None:
        use_log = logging.getLogger(__name__)
    use_log.info("Parameters")
    use_log.info("=" * 80)
    all_option_keys = set(kwargs.keys())
    all_option_keys |= set([x[0] for x in connection_option_keys])
    base_class = None

    if kwargs.get("flavour") == "amqp":
        base_class = Monkey
    elif kwargs.get("flavour") == "mqtt":
        base_class = Ape

    for key in sorted(all_option_keys):
        implementation_attribute = "DEFAULT_{:s}".format(key).upper()

        try:
            fallback_val = getattr(base_class, implementation_attribute)
        except AttributeError:
            fallback_val = None

        val = kwargs.get(key, fallback_val)
        use_log.info("{key:14}: {val!r}".format(key=key, val=val))


def hunchback_client():
    parser = argparse.ArgumentParser()
    flavours = ("amqp", "mqtt")
    connection_option_keys = (
        ("host", "a"),
        ("port", "i"),
        ("username", "u"),
        ("password", "p"),
    )

    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="Dry run mode",
    )

    publishing_group = parser.add_argument_group("Publishing")
    publishing_group.add_argument(
        "payloads",
        metavar="DATA",
        nargs="*",
        help="Payloads - either JSON encoded parameter or path of a JSON "
        "encoded file",
    )
    publishing_group.add_argument(
        "-r",
        "--routing-key",
        dest="routing_key",
        default="quasimodo.notifications",
        help="Topic",
    )

    flavour_group = parser.add_argument_group("Protocol")
    group = flavour_group.add_mutually_exclusive_group()
    for flavour in flavours:
        group.add_argument(
            "--{:s}".format(flavour),
            const=flavour,
            action="store_const",
            dest="flavour",
            default=flavours[0],
            help="Use {!r} as protocol ".format(flavour),
        )

    connection_group = parser.add_argument_group("Generic Connection Options")
    for opt_l, opt_s in connection_option_keys:
        env_key = "QUASIMODO_{:s}".format(opt_l).upper()
        connection_group.add_argument(
            "-{:s}".format(opt_s),
            "--{:s}".format(opt_l),
            dest=opt_l,
            default=os.environ.get(env_key, OPT_NO_VALUE),
            help="{!r} parameter. Default: %(default)s, "
            "Environment variable: {!r}".format(opt_l, env_key),
        )
    connection_group.add_argument(
        "--listen",
        dest="binding_keys",
        default=[],
        action="append",
        help="Subscriptions",
    )
    connection_group.add_argument(
        "--client_id",
        dest="client_id",
        default=None,
        help="Client ID to be used when connecting",
    )

    amqp_group = parser.add_argument_group("AMQP Connection Options")
    amqp_group_queue = amqp_group.add_mutually_exclusive_group()
    env_key_queue = "QUASIMODO_QUEUE"
    env_key_exchange = "QUASIMODO_EXCHANGE"
    amqp_group_queue.add_argument(
        "--queue",
        dest="queue",
        default=os.environ.get(env_key_queue, False),
        help="Queue, Environment variable: {!r}".format(env_key_queue),
    )
    amqp_group_queue.add_argument(
        "--exchange",
        dest="exchange",
        default=os.environ.get(env_key_exchange, "amq.topic"),
        help="Exchange, Environment variable: {!r}".format(env_key_exchange),
    )

    tls_group = parser.add_argument_group("TLS Options")
    tls_group.add_argument(
        "--no-tls",
        dest="tls",
        default=True,
        action="store_false",
        help="Disable TLS",
    )
    tls_group.add_argument(
        "--tls-ca",
        dest="cafile",
        default=None,
        help="Certificate Authority File Location",
    )
    tls_group.add_argument(
        "--tls-key",
        dest="keyfile",
        default=None,
        help="Private Key File Location",
    )
    tls_group.add_argument(
        "--tls-cert",
        dest="certfile",
        default=None,
        help="Public Key File Location",
    )
    tls_group.add_argument(
        "--transport-tcp",
        dest="transport",
        default=None,
        help="",
        const="tcp",
        action="store_const",
    )
    tls_group.add_argument(
        "--alpn-amazon",
        dest="alpn_amz",
        default=None,
        help="",
        action="store_true",
    )

    cli_args = parser.parse_args()
    kwargs = dict()

    for key, val in vars(cli_args).items():
        if val == OPT_NO_VALUE:
            continue
        elif key == "binding_keys" and not val:
            continue

        kwargs[key] = val

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y%m%d %H:%M:%S",
    )
    logging.getLogger("pika").setLevel(logging.WARNING)

    log = logging.getLogger(__name__)
    log.debug(kwargs)

    if cli_args.flavour == "amqp":
        impl = MonkeyDumper(**kwargs)
    elif cli_args.flavour == "mqtt":
        impl = ApeDumper(**kwargs)
    else:
        raise ValueError(
            "Implementation for {!r} is not available!".format(
                cli_args.flavour
            )
        )

    if cli_args.dry_run:
        log.info("DRY RUN.")
        log.info("")
        dump_hunchback_parameters(kwargs, connection_option_keys)
        sys.exit(0)

    payloads = []
    if cli_args.payloads:
        for p_in in cli_args.payloads:
            try:
                if os.path.isfile(p_in):
                    with open(p_in, "r") as src:
                        payload = json.load(src)
                else:
                    payload = json.loads(p_in)
                payloads.append(payload)
            except Exception as exc:
                log.info("Ignoring payload {!r}, GOT {!s}".format(p_in, exc))

        if not payloads:
            log.error("No payloads? No gain.")
            sys.exit(2)

    try:
        if payloads:
            log.info(payloads)
            for payload in payloads:
                impl.simple_publish(payload, cli_args.routing_key)
        else:
            impl.run()
    except socket.error as sexc:
        log.error("Could not connect: {!s}".format(sexc))
        dump_hunchback_parameters(kwargs, connection_option_keys)
        log.info("")
        log.info("Implementation Object:")
        log.info("=" * 80)
        log.info(impl)
        sys.exit(3)
    except KeyboardInterrupt:
        impl.log.info("You pressed Ctrl-C. This will be reported.")
        sys.exit(1)

    sys.exit(0)
